@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # vim:fenc=utf-8
 #
 # pylint: disable=too-many-arguments
@@ -10,6 +9,7 @@
 from collections import Counter
 import re
 import os
+import signal
 import sys
 import subprocess
 import logging
@@ -18,7 +18,6 @@ import time
 import datetime
 import configparser
 import glob
-import copy
 import shlex
 import shutil
 import ipaddress
@@ -38,7 +37,14 @@ SERVICE_OPTIONS_TYPE = {
     'ip_prefix': 'get',
     'interface': 'get',
     'ip_check_disabled': 'getboolean',
+    'custom_bird_reconfigure_cmd_timeout': 'getfloat',
+    'custom_bird_reconfigure_cmd': 'get',
 }
+SERVICE_OPTIONAL_OPTIONS = {
+    'custom_bird_reconfigure_cmd_timeout',
+    'custom_bird_reconfigure_cmd',
+}
+
 DAEMON_OPTIONS_TYPE = {
     'pidfile': 'get',
     'bird_conf': 'get',
@@ -62,12 +68,14 @@ DAEMON_OPTIONS_TYPE = {
     'bird6_changes_counter': 'getint',
     'bird_reconfigure_cmd': 'get',
     'bird6_reconfigure_cmd': 'get',
+    'splay_startup': 'getfloat',
 }
 DAEMON_OPTIONAL_OPTIONS = [
     'stderr_log_server',
     'stderr_file',
     'log_server',
     'log_file',
+    'splay_startup',
 ]
 
 
@@ -115,7 +123,7 @@ def touch(file_path):
 
 
 def get_ip_prefixes_from_config(config, services, ip_version):
-    """Build a set of IP prefixes found in configuration files.
+    """Build a set of IP prefixes found in service configuration files.
 
     Arguments:
         config (obg): A configparser object which holds our configuration.
@@ -168,10 +176,10 @@ def modify_ip_prefixes(
         ip_version):
     """Modify IP prefixes in Bird configuration.
 
-    - Depending on the configuration either remove or report IP prefixes found
+    Depending on the configuration either removes or reports IP prefixes found
     in Bird configuration for which we don't have a service check associated
-    with them.
-    - Add ``dummy_ip_prefix`` in Bird configuration if it is missing
+    with them. Moreover, it adds the dummy IP prefix if it isn't present and
+    ensures that the correct variable name is set.
 
     Arguments:
         config (obg): A configparser object which holds our configuration.
@@ -197,6 +205,16 @@ def modify_ip_prefixes(
                   "error, thus exiting main program", error)
         sys.exit(1)
 
+    _name = get_variable_name_from_bird(config_file)
+    if _name is None:
+        log.warning("failed to find variable name in %s, going to add it",
+                    config_file)
+        update_bird_conf = True
+    elif _name != variable_name:
+        log.warning("found incorrect variable name in %s, going to add the "
+                    "correct one %s", _name, variable_name)
+        update_bird_conf = True
+
     if dummy_ip_prefix not in ip_prefixes_in_bird:
         log.warning("dummy IP prefix %s is missing from bird configuration "
                     "%s, adding it", dummy_ip_prefix, config_file)
@@ -229,8 +247,6 @@ def modify_ip_prefixes(
                         ','.join(ip_prefixes_without_check),
                         config_file)
 
-    # Either dummy IP-Prefix was added or an IP-Prefix(es) without a service
-    # check was removed.
     if update_bird_conf:
         if keep_changes:
             archive_bird_conf(config_file, changes_counter)
@@ -270,15 +286,17 @@ def load_configuration(config_file, config_dir, service_file):
     Returns:
         A tuple with 1st element a ConfigParser object and 2nd element
         a dictionary.
+    Raises:
+        ValueError if a sanity check fails.
 
     """
     config_files = [config_file]
-    defaults = copy.copy(DEFAULT_OPTIONS['DEFAULT'])
-    daemon_defaults = {
-        'daemon': copy.copy(DEFAULT_OPTIONS['daemon'])
-    }
-    config = configparser.ConfigParser(defaults=defaults)
-    config.read_dict(daemon_defaults)
+    config = configparser.ConfigParser()
+    config.read_dict(DEFAULT_OPTIONS)
+
+    if not os.path.isfile(config_file):
+        raise ValueError("{f} configuration file either isn't readable or "
+                         "doesn't exist".format(f=config_file))
     if service_file is not None:
         if not os.path.isfile(service_file):
             raise ValueError("{f} configuration file for a service check "
@@ -376,6 +394,9 @@ def service_configuration_check(config):
         for option, getter in SERVICE_OPTIONS_TYPE.items():
             try:
                 getattr(config, getter)(service, option)
+            except configparser.NoOptionError as error:
+                if option not in SERVICE_OPTIONAL_OPTIONS:
+                    raise ValueError(error)
             except configparser.Error as error:
                 raise ValueError(error)
             except ValueError as exc:
@@ -498,6 +519,42 @@ def build_bird_configuration(config):
     return bird_configuration
 
 
+def get_variable_name_from_bird(bird_conf):
+    """Return the variable name set in Bird configuration.
+
+    The variable name in Bird configuration is set with the keyword 'define',
+    here is an example:
+
+        define ACAST_PS_ADVERTISE =
+
+    and we exract the string between the word 'define' and the equals sign.
+
+    Arguments:
+        bird_conf (str): The absolute file name path of Bird configuration.
+
+    Returns:
+        The variable name as a string or None if it isn't found.
+
+    """
+    bird_variable_pattern = re.compile(
+        r'''
+        ^\s*
+        define\s+
+        (?P<name>\S+\b)
+        \s+
+        =
+        ''', re.VERBOSE
+    )
+
+    with open(bird_conf, 'r') as content:
+        for line in content.readlines():
+            variable_match = bird_variable_pattern.search(line)
+            if variable_match:
+                return variable_match.group('name')
+
+    return None
+
+
 def create_bird_config_files(bird_configuration):
     """Create bird configuration files per IP version.
 
@@ -601,13 +658,26 @@ class BaseOperation(object):
     Arguments:
         name (string): The name of the service for the given ip_prefix
         ip_prefix (string): The value to run the operation
+        ip_version (string): IP protocol version
+        bird_reconfigure_cmd (string): A custom command to trigger
+        reconfiguration to Bird Daemon.
+        bird_reconfigure_cmd_timeout (float): Maximum time to wait for command
+        to complete.
     """
 
-    def __init__(self, name, ip_prefix, ip_version):  # noqa:D102
+    def __init__(
+            self,
+            name,
+            ip_prefix,
+            ip_version,
+            bird_reconfigure_cmd,
+            bird_reconfigure_timeout):  # noqa:D102
         self.name = name
         self.ip_prefix = ip_prefix
         self.log = logging.getLogger(PROGRAM_NAME)
         self.ip_version = ip_version
+        self.bird_reconfigure_cmd = bird_reconfigure_cmd
+        self.bird_reconfigure_timeout = bird_reconfigure_timeout
 
 
 class AddOperation(BaseOperation):
@@ -925,7 +995,7 @@ def setup_logger(config):
 
     custom_format = log_format()
     json_formatter = CustomJsonFormatter(custom_format,
-                                         prefix=PROGRAM_NAME)
+                                         prefix=PROGRAM_NAME + ': ')
     formatter = logging.Formatter(
         '%(asctime)s {program}[%(process)d] %(levelname)-8s '
         '%(threadName)-{width}s %(message)s'
@@ -1111,3 +1181,68 @@ class CustomJsonFormatter(jsonlogger.JsonFormatter):
         # return jsonlogger.JsonFormatter.process_log_record(self, log_record)
 
         return log_record
+
+
+class ServiceCheckDiedError(Exception):
+    """Raised when a thread that runs service check dies.
+
+    Arguments:
+        name (str): the name of the thread
+        trace (str): A trace, preferably the output of traceback.format_exc()
+    """
+
+    def __init__(self, name, trace):
+        """Initialize."""
+        super().__init__()
+        self.name = name
+        self.trace = trace
+
+    def __str__(self):
+        """More useful."""
+        return ("thread for service {n} died due to : {t}"
+                .format(n=self.name, t=self.trace))
+
+
+def run_custom_bird_reconfigure(operation):
+    """Reconfigure BIRD daemon by running a custom command.
+
+    It adds one argument to the command, either "up" or "down".
+    If command times out then we kill it. In order to avoid leaving any orphan
+    processes, that may have been started by the command, we start a new
+    session when we invoke the command and then we kill process group of that
+    session.
+
+    Arguments:
+        operation (obj): Either a AddOperation or DeleteOperation object.
+
+    """
+    log = logging.getLogger(PROGRAM_NAME)
+    if isinstance(operation, AddOperation):
+        status = 'up'
+    else:
+        status = 'down'
+    cmd = shlex.split(operation.bird_reconfigure_cmd + " " + status)
+    log.info("reconfiguring BIRD by running custom command %s", ' '.join(cmd))
+    try:
+        proc = subprocess.Popen(cmd,
+                                start_new_session=True,
+                                stdout=subprocess.PIPE,
+                                stderr=subprocess.PIPE)
+        _, errs = proc.communicate(
+            timeout=operation.bird_reconfigure_timeout
+        )
+    except OSError as exc:
+        log.error("reconfiguring BIRD failed with: %s", exc)
+    except subprocess.TimeoutExpired as exc:
+        log.error("reconfiguring bird timed out")
+        if proc.poll() is None:  # if process is still alive
+            try:
+                os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
+            except PermissionError as exc:
+                log.error("failed to terminate custom bird command: %s", exc)
+    else:
+        if proc.returncode != 0:
+            log.error("reconfiguring BIRD failed with return code: %s and "
+                      "stderr: %s", proc.returncode, errs)
+        else:
+            log.info("custom command successfully reconfigured Bird")
